//================================
//Exposed Functions.
//Intended for use by your code.
//================================

//If your event listener returns this, prevent the following listeners from being called.
//Value chosen arbitrary to avoid typographical errors.
const EVENT_CANCEL = 117115;

//You can skip `order` parameter. Default value is 0.
::OnGameEvent <- function(eventName, order, listenerFunc = null, scope = null, selfListener = false)
{
    if (endswith(eventName, "_post"))
        return OnGameEventPostInternal(eventName.slice(0, eventName.len() - 5), order, listenerFunc, scope, selfListener, 0);
    if (endswith(eventName, "_next"))
        return OnGameEventPostInternal(eventName.slice(0, eventName.len() - 5), order, listenerFunc, scope, selfListener, 0.01);
    else
        return OnGameEventInternal(eventName, order, listenerFunc, scope, selfListener, "OnGameEvent_", "GameEventCallbacks", RegisterScriptGameEventListener);
}

//Same as OnGameEvent, but fires ONLY if the event's player is `scope.self`.
//You can skip `order` parameter. Default value is 0.
::OnSelfEvent <- function(eventName, order, listenerFunc = null, scope = null)
{
    if (typeof(order) == "function")
    {
        scope = listenerFunc;
        listenerFunc = order;
        order = 0;
    }
    return OnGameEvent(eventName, order, listenerFunc, scope, true);
}

//==============================
//Internal Functions.
//==============================

::eventRootTable <- {};
eventRootTable.ScriptHookCallbacks <- {};
eventRootTable.GameEventCallbacks <- {};

::listenerTable <- {};

//The higher the value the LATER it will execute
::AddListener <- function(eventName, order, listenerFunc, scope = null, selfListener = false)
{
    if (scope == null)
        scope = this;
    else if ("IsValid" in scope)
    {
        scope.ValidateScriptScope();
        scope = scope.GetScriptScope();
    }
    local scope = scope.weakref();
    local parameters = listenerFunc.getinfos().parameters; //note: parameters[0] is hidden and it's always _scope_
    local paramLen = parameters.len();
    local hasPlayer = paramLen == 3 || (paramLen == 2 && (parameters[1] == "player" || parameters[1] == "victim")) ? 1 : 0;
    local hasParams = paramLen > (hasPlayer ? 2 : 1) ? 2 : 0;
    local listenerEntry = [order, listenerFunc, scope, hasPlayer + hasParams + (selfListener ? 4 : 0)];
    if (eventName in listenerTable)
    {
        local listenerQueue = listenerTable[eventName];
        local len = listenerQueue.len();
        local i = 0;
        for (; i < len; i++)
            if (listenerQueue[i][0] >= order)
                break;
        listenerQueue.insert(i, listenerEntry);
    }
    else
        listenerTable[eventName] <- [listenerEntry];
    return listenerEntry;
}

::FireListeners <- function(eventName, params)
{
    if (!(eventName in listenerTable))
        return;

    local player = null;
    if ("self" in params)
        player = self;
    else if ("userid" in params && eventName != "player_connect")
    {
        player = GetPlayerFromUserID(params.userid);
        if (!IsValidClient(player))
            return;
        if (eventName == "player_spawn" && params.team == 0)
        {
            ClientCache_Join(player);
            return;
        }
        if (eventName == "player_death" && params.death_flags & 32)
        {
            FireCustomEvent("player_death_feign", params);
            return;
        }
    }

    //Yes, we need to reevaluate the array length every step because scopes can be removed inside one of the listeners.
    local listenerQueue = listenerTable[eventName];
    for (local i = 0; i < listenerQueue.len(); i++)
    {
        local listenerEntry = listenerQueue[i];
        local scope = listenerEntry[2];

        if (!scope || ("self" in scope && (!scope.self || !scope.self.IsValid())))
        {
            listenerQueue.remove(i--);
            continue;
        }
        try
        {
            local flags = listenerEntry[3];
            if (flags & 4 && (!("self" in scope) || scope.self != player))
                continue;
            local args = [scope];
            if (flags & 1)
                args.push(player);
            if (flags & 2)
                args.push(params);
            local result = listenerEntry[1].acall(args);
            if (result == 117115)
                return;
        }
        catch (e) { } //This allows us to see the error in console, but it won't stop this cycle
    }
}
::FireCustomEvent <- FireListeners;

::OnGameEventInternal <- function(eventName, order, listenerFunc, scope, selfListener, prefix, eventType, regFunc)
{
    local eventTable = eventRootTable[eventType];
    if (!(eventName in eventTable))
    {
        regFunc(eventName);

        eventRootTable[prefix + eventName] <- function(params) { FireListeners(eventName, params); };
        eventTable[eventName] <- [];
        __CollectGameEventCallbacks(eventRootTable);
    }
    if (eventTable[eventName].find(main_script) == null)
        eventTable[eventName].append(main_script.weakref());
    return AddListener(eventName, order, listenerFunc, scope, selfListener);
}

::OnGameEventPostInternal <- function(eventName, order, listenerFunc, scope, selfListener, delay)
{
    local parameters = listenerFunc.getinfos().parameters;
    local paramLen = parameters.len();
    local hasPlayer = paramLen == 3 || (paramLen == 2 && (parameters[1] == "player" || parameters[1] == "victim" || parameters[1] == "self" || parameters[1] == "entity")) ? 1 : 0;
    local hasParams = paramLen > (hasPlayer ? 2 : 1) ? 2 : 0;

    return OnGameEvent(eventName, order, function(player, params) {
        local args = [this, delay, listenerFunc];
        if (hasPlayer)
            args.push(player);
        if (hasParams)
            args.push(params);
        args.push(this);
        RunWithDelay.acall(args);
    }, scope, selfListener);
}